/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 *
 * Guillaume Chazarain <gfc@altern.org>
 */

/************************************************
 * Creation and callbacks of the OpenGL widget. *
 ************************************************/

#include "gliv.h"

#include <GL/gl.h>
#include <gtk/gtkgl.h>

extern GtkWidget *gl_widget;
extern rt_struct *rt;
extern options_struct *options;

static gint old_width, old_height;
static gboolean hide_cursor_enabled = TRUE;

void update_bg_color()
{
    glClearColor(options->bg_col[0] / 65535.0, options->bg_col[1] / 65535.0,
                 options->bg_col[2] / 65535.0, 0.0);
}

gboolean toggle_cursor(gboolean visible)
{
    static GdkCursor *hidden_cursor = NULL;
    GdkCursor *new_cursor;

    if (hidden_cursor == NULL) {
        /* First time, create the invisible cursor. */
        GdkPixmap *pmap;
        GdkColor *col;
        gchar *data;

        data = g_new0(gchar, 1);
        col = g_new0(GdkColor, 1);
        pmap = gdk_bitmap_create_from_data(NULL, data, 1, 1);
        hidden_cursor = gdk_cursor_new_from_pixmap(pmap, pmap, col, col, 0, 0);

        g_free(data);
        g_free(col);
        gdk_bitmap_unref(pmap);
    }

    if (visible) {
        /* Show the cursor. */
        new_cursor = NULL;
        schedule_hide_cursor();
    } else {
        /* Hide the cursor. */
        new_cursor = hidden_cursor;
        gtk_widget_add_events(gl_widget, GDK_POINTER_MOTION_MASK);
    }

    rt->cursor_hidden = !visible;
    gdk_window_set_cursor(gl_widget->window, new_cursor);

    return FALSE;
}

/* Used to prevent the cursor from being toggled. */
void set_hide_cursor_enabled(gboolean enabled)
{
    if (enabled == hide_cursor_enabled)
        return;

    hide_cursor_enabled = enabled;
    if (hide_cursor_enabled == FALSE && rt->cursor_hidden)
        toggle_cursor(TRUE);

    schedule_hide_cursor();
}

void schedule_hide_cursor(void)
{
    static guint id = 0;

    if (id != 0)
        /* Remove previous schedule. */
        g_source_remove(id);

    if (options->delay == 0 || hide_cursor_enabled == FALSE)
        id = 0;
    else
        id = g_timeout_add(options->delay, (GSourceFunc) toggle_cursor,
                           GINT_TO_POINTER(FALSE));
}

static gboolean expose_event(GtkWidget * unused, GdkEventExpose * event)
{
    if (event->count == 0)
        refresh(REFRESH_IMAGE);

    return TRUE;
}

static void resize_widget(gint width, gint height)
{
    gdk_gl_drawable_wait_gdk(gtk_widget_get_gl_drawable(gl_widget));

    glViewport(0, 0, width, height);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    /* Vertical flip to have coordinates like X but centered like OpenGL. */
    glOrtho(-width / 2.0, width / 2.0, height / 2.0, -height / 2.0, -1.0, 1.0);

    glMatrixMode(GL_MODELVIEW);
}

static gboolean configure_event(GtkWidget * unused, GdkEventConfigure * event)
{
    gint what = REFRESH_IMAGE;

    resize_widget(event->width, event->height);

    if ((options->maximize || options->scaledown) &&
        matrix_set_max_zoom(old_width, old_height, FALSE) == FALSE) {
        /*
         * When options->maximize or options->scaledown is set we zoom only
         * if a matrix_set_max_zoom() with previous dimensions would have been
         * ignored.  It happens if the image properties didn't change after the
         * previous matrix_set_max_zoom().
         * This is achieved by the test of matrix_set_max_zoom().
         */
        matrix_set_max_zoom(event->width, event->height, TRUE);
        what |= REFRESH_STATUS;
    }

    refresh(what);

    old_width = event->width;
    old_height = event->height;

    return TRUE;
}

static void drag_data_received(GtkWidget * unused1, GdkDragContext * context,
                               gint unused2, gint unused3,
                               GtkSelectionData * data, guint unused4,
                               guint time)
{
    gchar *uri_list;
    gchar **uris;
    gint i;
    GError *err = NULL;

    /* Based on code from gnome-terminal. */

    if (data->format != 8) {
        ERROR_MSG(_("Wrong URI format: %d (instead of 8)\n"), data->format);
        return;
    }

    if (data->length <= 0) {
        ERROR_MSG(_("Wrong URI length: %d\n"), data->length);
        return;
    }

    uri_list = g_strndup(data->data, data->length);
    uris = g_strsplit(uri_list, "\r\n", -1);

    if (uris == NULL)
        return;

    for (i = 0; uris[i] != NULL; i++) {
        gchar *old = uris[i];

        if (old[0] == '\0')
            break;

        uris[i] = g_filename_from_uri(old, NULL, &err);
        if (err != NULL) {
            g_printerr("%s\n", err->message);
            g_error_free(err);
            err = NULL;
        }

        if (uris[i] == NULL)
            uris[i] = old;
        else
            g_free(old);
    }

    if (insert_after_current(uris, i, FALSE))
        open_next_image();

    g_strfreev(uris);
    g_free(uri_list);

    gtk_drag_finish(context, TRUE, TRUE, time);
}

static void dnd_init(void)
{
    GtkTargetEntry targets[] = {
/* *INDENT-OFF* */
        {"text/uri-list", 0, 0}
/* *INDENT-ON* */
    };

    gtk_drag_dest_set(gl_widget, GTK_DEST_DEFAULT_ALL, targets,
                      G_N_ELEMENTS(targets), GDK_ACTION_COPY);

    g_signal_connect(gl_widget, "drag-data-received",
                     G_CALLBACK(drag_data_received), NULL);
}

/*
 * Probe the max_texture_size with a GL_PROXY_TEXTURE_2D.
 * Used only when initializing the widget.
 */
static gboolean check_texture_size(void)
{
    gint width;

    glTexImage2D(GL_PROXY_TEXTURE_2D, 0, GL_RGB,
                 rt->max_texture_size, rt->max_texture_size, 0, GL_RGBA,
                 GL_UNSIGNED_BYTE, NULL);

    glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &width);

    return width == rt->max_texture_size;
}

static void get_max_texture_size(void)
{
    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &rt->max_texture_size);
    while (rt->max_texture_size > 0 && check_texture_size() == FALSE)
        rt->max_texture_size >>= 1;

    if (rt->max_texture_size < 64) {
        ERROR_MSG(_("Fatal error: GL_MAX_TEXTURE_SIZE = %d\n"),
                  rt->max_texture_size);
        ERROR_MSG(_("GL_MAX_TEXTURE_SIZE must be >= 64\n"));
        quit(1);
    }
}

static gboolean make_current(void)
{
    GdkGLContext *glcontext = gtk_widget_get_gl_context(gl_widget);
    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(gl_widget);

    return gdk_gl_drawable_make_current(gldrawable, glcontext);
}

static gboolean first_expose(void)
{
    /* We only handle the first one. */
    g_signal_handlers_disconnect_by_func(gl_widget, G_CALLBACK(first_expose),
                                         NULL);

    if (make_current() == FALSE) {
        ERROR_MSG(_("glXMakeCurrent() failed\n"));
        quit(1);
    }

    glDisable(GL_DEPTH_TEST);

    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glShadeModel(GL_FLAT);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    update_bg_color();

    get_max_texture_size();

    rt->wid_size = &gl_widget->allocation;

    g_signal_connect(gl_widget, "configure-event",
                     G_CALLBACK(configure_event), NULL);

    g_signal_connect(gl_widget, "expose-event", G_CALLBACK(expose_event), NULL);

    schedule_hide_cursor();

    old_width = rt->wid_size->width;
    old_height = rt->wid_size->height;

    if (options->fullscreen)
        toggle_fullscreen(TRUE);

    resize_widget(rt->wid_size->width, rt->wid_size->height);

    dnd_init();

    refresh_all();
    load_first_image();

    refresh_all();
    load_second_image();
    refresh_all();

    if (options->build_menus)
        rebuild_images_menus();

    return TRUE;
}

void create_gl_widget(void)
{
    GdkGLConfigMode mode = GDK_GL_MODE_RGBA | GDK_GL_MODE_DEPTH |
        GDK_GL_MODE_DOUBLE | GDK_GL_MODE_ALPHA;

    GdkGLConfig *glconfig;

    if (gdk_gl_query_extension() == FALSE) {
        ERROR_MSG(_("OpenGL not supported\n"));
        quit(1);
    }

    glconfig = gdk_gl_config_new_by_mode(mode);

    if (glconfig == NULL)
        /* Try without the alpha channel. */
        glconfig = gdk_gl_config_new_by_mode(mode & ~GDK_GL_MODE_ALPHA);

    if (glconfig == NULL) {
        ERROR_MSG(_("Cannot find an appropriate visual, try glxinfo(1)\n"));
        quit(1);
    }

    gl_widget = gtk_drawing_area_new();
    if (gtk_widget_set_gl_capability(gl_widget, glconfig, NULL, TRUE,
                                     GDK_GL_RGBA_TYPE) == FALSE) {
        ERROR_MSG(_("Cannot set the OpenGL capability\n"));
        quit(1);
    }

    gtk_widget_set_events(gl_widget,
                          GDK_BUTTON_PRESS_MASK | GDK_BUTTON1_MOTION_MASK |
                          GDK_BUTTON3_MOTION_MASK | GDK_BUTTON_RELEASE_MASK);

    if (options->delay != 0)
        gtk_widget_add_events(gl_widget, GDK_POINTER_MOTION_MASK);

    /* The real initialization is done on the first expose. */
    g_signal_connect_after(gl_widget, "expose-event", G_CALLBACK(first_expose),
                           NULL);
}
