/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 *
 * Guillaume Chazarain <gfc@altern.org>
 */

/*********************
 * Main source file. *
 *********************/

#include "gliv.h"
#include "cmdline.h"

#include <stdlib.h>             /* exit() */
#include <unistd.h>             /* close(), dup(), dup2(), STDERR_FILENO */
#include <gtk/gtkgl.h>

rt_struct *rt;
options_struct *options;
gliv_image *current_image = NULL;
GtkWidget *gl_widget;
GtkMenuBar *menu_bar;

/*
 * This is borrowed from gtk+-2, without the gtk_grab_notify() calls.
 * They slow down things a lot when the images menus are built with many files.
 */
GtkWindowGroup *_gtk_window_get_group(GtkWindow * window)
{
    if (window && window->group)
        return window->group;
    else {
        static GtkWindowGroup *default_group = NULL;

        if (!default_group)
            default_group = gtk_window_group_new();

        return default_group;
    }
}

static GtkWindowGroup *gtk_main_get_window_group(GtkWidget * widget)
{
    GtkWidget *toplevel = NULL;

    if (widget)
        toplevel = gtk_widget_get_toplevel(widget);

    if (toplevel && GTK_IS_WINDOW(toplevel))
        return _gtk_window_get_group(GTK_WINDOW(toplevel));
    else
        return _gtk_window_get_group(NULL);
}

void gtk_grab_add(GtkWidget * widget)
{
    GtkWindowGroup *group;

    g_return_if_fail(widget != NULL);

    if (!GTK_WIDGET_HAS_GRAB(widget) && GTK_WIDGET_IS_SENSITIVE(widget)) {
        GTK_WIDGET_SET_FLAGS(widget, GTK_HAS_GRAB);

        group = gtk_main_get_window_group(widget);

        g_object_ref(widget);
        group->grabs = g_slist_prepend(group->grabs, widget);

/*      gtk_grab_notify (group, widget, FALSE); */
    }
}

void gtk_grab_remove(GtkWidget * widget)
{
    GtkWindowGroup *group;

    g_return_if_fail(widget != NULL);

    if (GTK_WIDGET_HAS_GRAB(widget)) {
        GTK_WIDGET_UNSET_FLAGS(widget, GTK_HAS_GRAB);

        group = gtk_main_get_window_group(widget);
        group->grabs = g_slist_remove(group->grabs, widget);

        g_object_unref(widget);

/*      gtk_grab_notify (group, widget, TRUE); */
    }
}

#define check_flag(given, flag, value) if (given) flag = value;

typedef struct {
    gboolean given;
    gboolean *flag;
    gboolean value;
} ggo_arg;

static void fill_flags(struct gengetopt_args_info *ggo)
{
/* *INDENT-OFF* */
    ggo_arg args[] = {
{ ggo->zoom_pointer_given, &options->zoom_pointer, ggo->zoom_pointer_flag },
{ ggo->dither_given,       &options->dither,       ggo->dither_flag       },
{ ggo->force_load_given,   &options->force,        ggo->force_load_flag   },
{ ggo->mipmap_given,       &options->mipmap,       ggo->mipmap_flag       },
{ ggo->build_menus_given,  &options->build_menus,  ggo->build_menus_flag  },
{ ggo->mnemonics_given,    &options->mnemonics,    ggo->mnemonics_flag    },
{ ggo->menu_given,         &options->menu_bar,     ggo->menu_flag         },
{ ggo->info_given,         &options->status_bar,   ggo->info_flag         },
{ ggo->scrollbars_given,   &options->scrollbars,   ggo->scrollbars_flag   },
{ ggo->alpha_checks_given, &options->alpha_checks, ggo->alpha_checks_flag },
{ ggo->loop_given,         &options->loop,         ggo->loop_flag         },
{ FALSE,                   NULL,                   FALSE                  }
};
/* *INDENT-ON* */

    ggo_arg *arg;

    for (arg = args; arg->flag != NULL; arg++)
        check_flag(arg->given, *(arg->flag), arg->value);
}

static void init_args(gint argc, gchar ** argv)
{
    struct gengetopt_args_info ggo;
    options_struct *rc_file;

    /* Command line (only the glivrc flag). */
    if (cmdline_parser(argc, argv, &ggo) != 0)
        quit(1);

    /* Configuration file. */
    rc_file = load_rc(ggo.glivrc_flag);
    options = g_memdup(rc_file, sizeof(options_struct));

    /* Command line (remaining options). */

    if (ggo.full_screen_given || ggo.make_fit_given ||
        ggo.maximize_given || ggo.scale_down_given) {

        options->fullscreen = ggo.full_screen_flag || ggo.make_fit_flag ||
            ggo.maximize_flag || ggo.scale_down_flag;

        if (ggo.make_fit_given)
            options->maximize = options->scaledown = ggo.make_fit_flag;
        else {

            check_flag(ggo.maximize_given, options->maximize,
                       ggo.maximize_flag);

            check_flag(ggo.scale_down_given, options->scaledown,
                       ggo.scale_down_flag);
        }

    } else if (options->maximize || options->scaledown)
        options->fullscreen = TRUE;

    if (ggo.delay_given && ggo.delay_arg >= 0)
        options->delay = ggo.delay_arg;

    if (ggo.history_given && ggo.history_arg >= -1)
        options->history_size = ggo.history_arg;

    if (ggo.duration_given && ggo.duration_arg >= 0)
        options->duration = ggo.duration_arg;

    fill_flags(&ggo);

    rt = g_new(rt_struct, 1);

    rt->cursor_hidden = FALSE;
    rt->help = FALSE;
    rt->alpha_checks_changed = TRUE;
    rt->scr_width = gdk_screen_width();
    rt->scr_height = gdk_screen_height();

    if (ggo.inputs_num > 0) {
        /* There are filenames on the command line. */
        init_list(ggo.inputs, ggo.inputs_num,
                  ggo.recursive_flag, ggo.shuffle_given, ggo.sort_given);

        if (ggo.slide_show_flag)
            begin_with_slide_show();
    }

    g_free(ggo.inputs);
}

/*
 * We call cmdline_parser() a first time to quickly exit if --help or --version
 * was given.  We do the real work after the next call in init_args(), after
 * gtk_init() removed the gtk arguments.
 *
 * We temporarily close stderr because unknown and gtk arguments are handled
 * afterwards.
 */
static void check_quick_exit(gint argc, gchar ** argv)
{
    struct gengetopt_args_info ggo;
    gint fd;

    fd = dup(STDERR_FILENO);
    close(STDERR_FILENO);

    cmdline_parser(argc, argv, &ggo);

    dup2(fd, STDERR_FILENO);
    close(fd);
}

G_GNUC_NORETURN void quit(gint code)
{
    imagemagick_finish();
    exit(code);
}

gint main(gint argc, gchar ** argv)
{
    /* i18n */
#ifdef ENABLE_NLS
    gtk_set_locale();
#   ifdef LOCALEDIR
    bindtextdomain(PACKAGE, LOCALEDIR);
#   endif
    textdomain(PACKAGE);
    bind_textdomain_codeset(PACKAGE, "UTF-8");
#endif

    check_quick_exit(argc, argv);

    gtk_init(&argc, &argv);
    gtk_gl_init(&argc, &argv);

    init_args(argc, argv);

    create_windows();

    gtk_main();

    return 0;
}
