/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 *
 * Guillaume Chazarain <gfc@altern.org>
 */

/*************************************
 * The Images and Directories menus. *
 *************************************/

#include "gliv.h"

#include <string.h>             /* strlen(), strchr(), strrchr() */

extern options_struct *options;

static gchar **array = NULL;
static gint length;
static GtkMenuItem *rebuilding_entry;

/* Rebuilding a menu is not reentrant. */
static gboolean rebuilding = FALSE;

/*** Functions for both menus. ***/

void set_rebuilding_entry(GtkMenuItem * item)
{
    rebuilding_entry = item;
}

static void set_menu_indicator(gchar * name, gint percent)
{
    static gchar *rebuilding = NULL;

    if (name == NULL)
        set_menu_label(rebuilding_entry, "", FALSE);
    else {
        gchar *label;

        if (rebuilding == NULL)
            /* First time. */
            rebuilding = _("Rebuilding:");

        label = g_strdup_printf("%s %s (%d%%)", rebuilding, name, percent);
        set_menu_label(rebuilding_entry, label, FALSE);

        g_free(label);
    }
}

static void set_progress(gchar * menu, gint * percent, gint file_id)
{
    if ((*percent + 1) * length <= 100 * file_id) {
        *percent = (100 * file_id) / length;
        set_menu_indicator(menu, *percent);
    }
}

/* We don't want to perform the signal lookup for each file. */
static void connect_activate(GtkMenuItem * instance, const gchar * filename)
{
    static guint signal_id = 0;
    GClosure *closure;

    if (signal_id == 0)
        /* First time. */
        signal_id = g_signal_lookup("activate", G_TYPE_FROM_INSTANCE(instance));

    closure = g_cclosure_new_swap(G_CALLBACK(menu_load),
                                  (gpointer) filename, NULL);

    g_signal_connect_closure_by_id(instance, signal_id, 0, closure, FALSE);
}

static void add_menu_item(GtkMenuShell * menu,
                          const gchar * name, const gchar * load)
{
    GtkMenuItem *item;

    name = locale_to_utf8(name);

    if (options->mnemonics && load != NULL) {
        name = add_mnemonic(name);
        item = GTK_MENU_ITEM(gtk_menu_item_new_with_mnemonic(name));
    } else
        item = GTK_MENU_ITEM(gtk_menu_item_new_with_label(name));

    gtk_menu_shell_insert(menu, GTK_WIDGET(item), 1);

    if (load == NULL)
        /* A dirname */
        gtk_widget_set_sensitive(GTK_WIDGET(item), FALSE);
    else
        /* A filename */
        connect_activate(item, load);

    process_events();
}

/*
 * Underscores in filenames are replaced
 * with mnemonics, so we duplicate them.
 *
 * The caller has to determine whether the
 * returned string has to be freed.
 */
static gchar *duplicate_underscores(gchar * orig)
{
    gint length = 0;
    gint nb_underscores = 0;
    const gchar *ptr_orig;
    gchar *new, *ptr_new;

    if (options->mnemonics == FALSE)
        return orig;

    /* How many underscores? */
    for (ptr_orig = orig; *ptr_orig != '\0'; ptr_orig++) {
        length++;
        if (*ptr_orig == '_')
            nb_underscores++;
    }

    if (nb_underscores == 0)
        return orig;

    ptr_new = new = g_new(gchar, length + nb_underscores + 1);

    for (ptr_orig = orig; *ptr_orig != '\0'; ptr_orig++) {
        *ptr_new = *ptr_orig;
        if (*ptr_orig == '_') {
            /* Duplicate this one. */
            ptr_new++;
            *ptr_new = '_';
        }
        ptr_new++;
    }

    *ptr_new = '\0';

    return new;
}

typedef enum {
    DIR_PARENT,
    DIR_EQUAL,
    DIR_DIFFERENT
} dir_type;

/*
 * Whether dir1 is parent, equal or different from the dir where file2 is.
 * wrt: with respect to.
 */
static dir_type dir_wrt_file(const gchar * dir1, const gchar * file2)
{
    const gchar *ptr1;
    gchar *ptr2, *last_slash;
    dir_type res;

    if (dir1[0] == '\0')
        /* dir1 == "". */
        return DIR_PARENT;

    if (dir1[0] == '/' && dir1[1] == '\0' && file2[0] == '/') {
        /* dir1 == "/" and file2 == "/...". */
        return strchr(file2 + 1, '/') == NULL ? DIR_EQUAL : DIR_PARENT;
    }

    /* We temporarily cut the filename to have the dirname. */
    ptr2 = (gchar *) file2;
    last_slash = strrchr(ptr2, '/');
    if (last_slash)
        *last_slash = '\0';

    ptr1 = dir1;

    /* Skip identical characters in the paths. */
    while (*ptr1 != '\0' && *ptr1 == *ptr2) {
        ptr1++;
        ptr2++;
    }

    if (*ptr1 == *ptr2)
        res = DIR_EQUAL;

    else if (*ptr1 == '\0' && *ptr2 == '/')
        res = DIR_PARENT;

    else
        res = DIR_DIFFERENT;

    if (last_slash)
        *last_slash = '/';
    return res;
}

/*** Directories menu. ***/

/* When dir == "parent/a/b/c...", it becomes "a". */
static gchar *first_dir(const gchar * dir, const gchar * parent)
{
    const gchar *res;

    if (parent[0] == '\0' && dir[0] == '/')
        return g_strdup("/");

    /* Skip identical characters in the paths. */
    while (*dir == *parent) {
        dir++;
        parent++;
    }

    /* Suppress the leading separator. */
    if (dir[0] == '/')
        dir++;

    res = dir;

    /* Keep only the first dir by looking for the second one, */
    while (*dir != '\0' && *dir != '/')
        dir++;

    /* and throwing it away. */
    return g_strndup(res, dir - res);
}

static void make_tree_rec(GNode * parent, gint * file_id,
                          const gchar * current_dir)
{
    gchar *this_dir, *tmp, *tmp_dir;
    GNode *node;

    this_dir = (current_dir[0] == '\0') ? g_strdup(parent->data) :
        g_build_filename(current_dir, parent->data, NULL);

    while (*file_id < length) {

        switch (dir_wrt_file(this_dir, array[*file_id])) {
        case DIR_PARENT:
            /* Subdirectory => add it, then recursive call. */
            tmp_dir = g_path_get_dirname(array[*file_id]);
            tmp = first_dir(tmp_dir, this_dir);

            node = g_node_new(tmp);
            g_node_append(parent, node);

            if (tmp[0] != '/' || tmp[1] != '\0')
                /* tmp != "/" */
                tmp = this_dir;

            make_tree_rec(node, file_id, tmp);
            break;

        case DIR_EQUAL:
            /* Same directory => add the filename, stay in the loop. */
            tmp = array[*file_id] + strlen(this_dir);
            if (tmp[0] == '/')
                tmp++;

            g_node_append_data(parent, g_strdup(tmp));
            (*file_id)++;
            break;

            /* case DIR_DIFFERENT: */
        default:
            /* Return from the recursive calls to find a common directory. */
            g_free(this_dir);
            return;
        }
    }

    g_free(this_dir);
}

static void add_file_from_tree(GtkMenuShell * menu, GNode * leaf)
{
    static gchar *directories = NULL;
    static gint file_id = 0, percent = 0;
    gchar *load, *menu_name, *tmp;
    GNode *node;

    if (directories == NULL)
        /* First time. */
        directories = _("Directories");

    if (menu == NULL) {
        file_id = 0;
        percent = 0;
        return;
    }

    load = g_strdup("");
    menu_name = duplicate_underscores(leaf->data);

    /* Rebuild the filename using the tree. */
    for (node = leaf; node != NULL; node = node->parent)
        /* Skip nodes containing only "". */
        if (((gchar *) (node->data))[0] != '\0') {
            tmp = load;

            load = fast_build_filename(node->data, load);

            g_free(tmp);
        }

    add_menu_item(menu, menu_name, load);

    if (leaf->data != menu_name)
        g_free(menu_name);

    set_progress(directories, &percent, file_id);
    file_id++;
}

static GtkMenuShell *add_menu(GtkMenuShell * parent, const gchar * name)
{
    GtkMenuShell *menu;
    GtkMenuItem *item;

    name = locale_to_utf8(name);

    if (options->mnemonics) {
        name = add_mnemonic(name);
        item = GTK_MENU_ITEM(gtk_menu_item_new_with_mnemonic(name));
    } else
        item = GTK_MENU_ITEM(gtk_menu_item_new_with_label(name));

    gtk_menu_shell_insert(parent, GTK_WIDGET(item), 1);

    menu = GTK_MENU_SHELL(gtk_menu_new());
    gtk_menu_item_set_submenu(item, GTK_WIDGET(menu));

    gtk_menu_shell_prepend(menu, gtk_tearoff_menu_item_new());

    return menu;
}

static void make_menu_from_tree_rec(GNode * tree, GtkMenuShell * parent)
{
    GtkMenuShell *menu;

    while (tree->children != NULL && tree->children->next == NULL) {
        /*
         * Only one child.
         * Compress 'tree->child' to '""->tree/child'.
         */
        gchar *dir;
        GNode *child;

        child = g_node_first_child(tree);
        if (((gchar *) (tree->data))[0] != '\0') {

            dir = fast_build_filename(tree->data, child->data);

            g_free(tree->data);
            g_free(child->data);

            tree->data = g_strdup("");
            child->data = dir;
        }
        tree = child;
    }

    if (G_NODE_IS_LEAF(tree))
        add_file_from_tree(parent, tree);
    else {
        menu = add_menu(parent, tree->data);
        g_node_children_foreach(tree, G_TRAVERSE_ALL,
                                (GNodeForeachFunc) make_menu_from_tree_rec,
                                menu);
    }
}

/*** Images menu. ***/

static void add_file_item(GtkMenuShell * menu, gint file_id)
{
    gchar *filename, *basename, *menu_name;

    filename = array[file_id];
    basename = strrchr(filename, '/') + 1;
    menu_name = duplicate_underscores(basename);

    add_menu_item(menu, menu_name, filename);

    if (menu_name != basename)
        g_free(menu_name);
}

/*** Menus builders. ***/

static GtkWidget *common_rebuild(GtkMenuItem * root, gboolean start)
{
    static GtkMenuShell *menu;
    static gboolean destroy_array;

    if (start && (rebuilding || is_loading()))
        /*
         * There is no problem in rebuilding the images menus during
         * a loading, it would just make the loading too long.
         */
        return NULL;

    rebuilding = start;
    if (root != NULL)
        gtk_menu_item_deselect(root);

    if (start == FALSE) {
        const gchar *label;
        GtkMenuItem *item;

        if (destroy_array) {
            g_free(array);
            array = NULL;
        }

        label = add_mnemonic(_("Rebuild this menu"));
        item = GTK_MENU_ITEM(gtk_menu_item_new_with_mnemonic(label));

        gtk_menu_shell_insert(menu, GTK_WIDGET(item), 1);

        return GTK_WIDGET(item);
    }

    if (array == NULL) {
        array = get_sorted_files_array();
        length = get_list_length();

        if (array == NULL)
            return NULL;

        destroy_array = TRUE;
    } else
        /* Called by rebuild_images_menus(). */
        destroy_array = FALSE;

    menu = GTK_MENU_SHELL(gtk_menu_new());
    gtk_menu_shell_prepend(menu, gtk_tearoff_menu_item_new());
    gtk_menu_item_set_submenu(root, GTK_WIDGET(menu));

    return GTK_WIDGET(menu);
}

static gboolean traverse_free(GNode * node, gpointer unused)
{
    g_free(node->data);
    return FALSE;
}

static void lower_root(GNode * tree)
{
    GNode *node;
    gchar *path;

    for (node = tree->children; node != NULL; node = node->next) {
        path = g_build_filename(tree->data, node->data, NULL);
        g_free(node->data);
        node->data = path;
    }
}

#define REBUILD_START(name)                                                  \
    do {                                                                     \
        if (root != NULL) {                                                  \
            item = root;                                                     \
            return;                                                          \
        }                                                                    \
                                                                             \
        menu = GTK_MENU_SHELL(common_rebuild(item, TRUE));                   \
        if (menu == NULL)                                                    \
            return;                                                          \
                                                                             \
        set_menu_indicator(name, 0);                                         \
    } while (0)

#define REBUILD_END(function)                                                \
    do {                                                                     \
        GtkMenuItem *launcher;                                               \
                                                                             \
        launcher = GTK_MENU_ITEM(common_rebuild(NULL, FALSE));               \
        g_signal_connect_swapped(launcher, "activate", G_CALLBACK(function), \
                                 NULL);                                      \
                                                                             \
        gtk_widget_show_all(GTK_WIDGET(item));                               \
        set_menu_indicator(NULL, 0);                                         \
    } while (0)

void rebuild_directories(GtkMenuItem * root)
{
    static GtkMenuItem *item;
    GtkMenuShell *menu;
    GNode *tree, *node;
    gint file_id = 0;

    REBUILD_START(_("Directories"));

    /* Build the tree. */
    tree = g_node_new(g_strdup(""));
    while (file_id < length) {
        make_tree_rec(tree, &file_id, tree->data);
    }

    /* Build the menu. */

    add_file_from_tree(NULL, NULL);

    if (tree->children != NULL && tree->children->next == NULL) {
        /* Only one child */
        gchar *path;

        node = tree->children;
        path = node->data;

        if ((path[0] == '.' || path[0] == '/') && path[1] == '\0')
            lower_root(node);
    } else
        node = tree;

    for (node = node->children; node != NULL; node = node->next)
        make_menu_from_tree_rec(node, menu);

    /* Destroy the tree. */
    g_node_traverse(tree, G_PRE_ORDER, G_TRAVERSE_ALL, -1, traverse_free, NULL);
    g_node_destroy(tree);

    REBUILD_END(rebuild_directories);
}

void rebuild_images(GtkMenuItem * root)
{
    static GtkMenuItem *item;
    gchar *menu_name;
    GtkMenuShell *menu;
    gint file_id;
    gchar *dir;
    gint percent = 0;

    menu_name = _("Images");
    REBUILD_START(menu_name);

    /* The file_id == 0 case. */
    dir = g_path_get_dirname(array[0]);
    add_file_item(menu, 0);

    for (file_id = 1; file_id < length; file_id++) {

        set_progress(menu_name, &percent, file_id);

        if (dir_wrt_file(dir, array[file_id]) != DIR_EQUAL) {
            /* We add the previous dir, since the array is reversed. */
            add_menu_item(menu, dir, NULL);

            g_free(dir);
            dir = g_path_get_dirname(array[file_id]);
        }

        add_file_item(menu, file_id);
    }

    add_menu_item(menu, dir, NULL);
    REBUILD_END(rebuild_images);
}

void rebuild_images_menus(void)
{
    if (rebuilding || is_loading())
        return;

    length = get_list_length();
    if (length == 0)
        return;

    array = get_sorted_files_array();

    rebuild_directories(NULL);
    rebuild_images(NULL);

    g_free(array);
    array = NULL;
}
