/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 *
 * Guillaume Chazarain <gfc@altern.org>
 */

/************************************
 * Loading and unloading of images. *
 ************************************/

#include "gliv.h"

#include <stdio.h>              /* FILE, fopen(), fread(), fclose(), perror() */
#include <signal.h>             /* signal(), SIG_DFL, SIGSEGV */
#include <string.h>             /* strlen(), strrchr() */
#include <GL/gl.h>

extern rt_struct *rt;
extern options_struct *options;

static const gchar *loading_filename = NULL;

gboolean is_loading(void)
{
    return loading_filename != NULL;
}

void destroy_image(gliv_image * im)
{
    gint level;
    texture_map *map;

    if (im != NULL) {
        for (level = 0; level < im->nb_maps; level++) {
            map = im->maps + level;

            glDeleteTextures(map->nb_tiles, map->tex_ids);
            g_free(map->tex_ids);

            glDeleteLists(map->list, map->nb_tiles);
            g_free(map->tiles);
        }

        g_free(im->maps);
        g_free(im->ident);
        g_free(im);
    }
}

static void segv_handler(gint sig)
{
    const gchar *sig_str;

    signal(sig, SIG_DFL);
    sig_str = locale_to_utf8(g_strsignal(sig));

    if (loading_filename == NULL)
        ERROR_MSG(_("%s not while loading an image\n"), sig_str);
    else
        ERROR_MSG(_("%s while loading %s\n"),
                  sig_str, locale_to_utf8(loading_filename));

    imagemagick_finish();

    raise(sig);
}

/* Called when the first image is loaded. */
void install_segv_handler(void)
{
    signal(SIGSEGV, segv_handler);
}

void fill_ident(gliv_image * im)
{
    gchar *alpha_str;

    g_free(im->ident);

    alpha_str = im->has_alpha ? _("alpha channel") : _("no alpha channel");

    im->ident = g_strdup_printf("%s (%u/%u) (%s)",
                                locale_to_utf8(im->node->data),
                                im->number + 1, get_list_length(), alpha_str);
}

static void fast_downcase(gchar * str)
{
    while (*str) {
        if (*str >= 'A' && *str <= 'Z')
            *str += 'a' - 'A';

        str++;
    }
}

static gboolean is_decompressible(const gchar * filename, gint ext_len)
{
    const struct format *res;

    /* We want the image extension between image_ext_start and image_ext_end. */
    const gchar *image_ext_end = filename + strlen(filename) - ext_len - 1;
    const gchar *image_ext_start = image_ext_end - 1;
    gint image_ext_length = 0;
    gchar *image_ext;

    while (image_ext_start > filename && *image_ext_start != '.') {
        image_ext_length++;
        image_ext_start--;
    }

    image_ext = g_strndup(image_ext_start + 1, image_ext_length);
    fast_downcase(image_ext);
    res = ext_to_loader(image_ext, image_ext_length);

    g_free(image_ext);

    if (res == NULL) {
        ERROR_MSG(_("%s: unknown decompressed extension\n"),
                  locale_to_utf8(filename));
        return FALSE;
    }

    if (res->loader == LOADER_DECOMP) {
        ERROR_MSG(_("%s: image cannot be compressed twice\n"),
                  locale_to_utf8(filename));
        return FALSE;
    }

    return TRUE;
}

gboolean is_loadable(const gchar * filename)
{
    const struct format *res;
    gchar *ext;
    gint ext_len;

    if (options->force)
        return TRUE;

    ext = strrchr(filename, '.');
    if (ext == NULL) {
        ERROR_MSG(_("%s: unknown extension  (none)\n"),
                  locale_to_utf8(filename));
        return FALSE;
    }

    ext = g_strdup(ext + 1);

    /* Downcase the extension so that we can do case sensitive strcmp. */
    fast_downcase(ext);

    ext_len = strlen(ext);
    res = ext_to_loader(ext, ext_len);

    g_free(ext);

    if (res == NULL) {
        ERROR_MSG(_("%s: unknown extension\n"), locale_to_utf8(filename));
        return FALSE;
    }

    if (res->loader == LOADER_DECOMP)
        return is_decompressible(filename, ext_len);

    return TRUE;
}

#if 0
/*
 * We used to blacken image pixels with alpha == 0.
 * These pixels are invisible but may alter the bilinear filtering.
 * We have stopped doing it since the time lost is too big for the benefit.
 */
static void clean_alpha(GdkPixbuf * pixbuf)
{
    guchar *pixel, *end;

    process_events();

    pixel = gdk_pixbuf_get_pixels(pixbuf);
    end = pixel + gdk_pixbuf_get_rowstride(pixbuf) *
        gdk_pixbuf_get_height(pixbuf);

    for (; pixel < end; pixel += 4) {
        if (pixel[3] == 0)
            /*
             * alpha == 0
             * This is not valid C since a cast cannot be a lvalue but...
             */
            *((guint32 *) pixel) = 0;
    }
}
#endif

static void set_loading(const gchar * filename)
{
    /* Used if there is a segfault. */
    loading_filename = filename;

    set_loading_entry(filename);
}

GdkPixbuf *load_gdk_pixbuf(const gchar * filename)
{
    static guchar buf[4096];
    gsize size;
    FILE *file;
    GdkPixbufLoader *loader;
    GdkPixbuf *pixbuf;
    GError *err = NULL;

    file = fopen(filename, "r");
    if (file == NULL) {
        perror(filename);
        return NULL;
    }

    loader = gdk_pixbuf_loader_new();

    while ((size = fread(buf, 1, sizeof(buf), file)) > 0) {
        if (gdk_pixbuf_loader_write(loader, buf, size, &err) == FALSE) {
            g_printerr("%s: %s\n", locale_to_utf8(filename), err->message);
            g_error_free(err);
            err = NULL;
            break;
        }

        process_events();
    }

    gdk_pixbuf_loader_close(loader, &err);
    if (err != NULL) {
        g_printerr("%s: %s\n", locale_to_utf8(filename), err->message);
        g_error_free(err);
    }

    pixbuf = gdk_pixbuf_loader_get_pixbuf(loader);
    if (pixbuf != NULL)
        gdk_pixbuf_ref(pixbuf);

    fclose(file);
    g_object_unref(loader);
    return pixbuf;
}

static GdkPixbuf *file_to_pixbuf(const gchar * filename)
{
    GdkPixbuf *pixbuf;

    pixbuf = load_gdk_pixbuf(filename);

    if (pixbuf == NULL)
        pixbuf = load_with_imagemagick(filename);

    if (pixbuf == NULL)
        ERROR_MSG(_("Cannot load %s\n"), locale_to_utf8(filename));

    return pixbuf;
}

G_GNUC_PURE static gint nb_maps(gint dim)
{
    gint ret = 1;

    while (dim > rt->max_texture_size) {
        dim *= MIPMAP_RATIO;
        ret++;
    }

    return ret;
}

gliv_image *load_file(const gchar * filename)
{
    gliv_image *im;
    GdkPixbuf *loaded_image;

    set_loading(filename);
    process_events();

    loaded_image = file_to_pixbuf(filename);
    if (loaded_image == NULL) {
        /* Loading failed. */
        set_loading(NULL);
        return NULL;
    }

    im = g_new(gliv_image, 1);
    im->ident = NULL;

    im->width = gdk_pixbuf_get_width(loaded_image);
    im->height = gdk_pixbuf_get_height(loaded_image);
    im->has_alpha = gdk_pixbuf_get_has_alpha(loaded_image);

    if (options->mipmap)
        im->nb_maps = MIN(nb_maps(im->width), nb_maps(im->height));
    else
        im->nb_maps = 1;

    im->maps = g_new(texture_map, im->nb_maps);
    im->maps[0].map_init = g_new(loading_data, 1);
    im->maps[0].map_init->pixbuf = loaded_image;

#if 0
    if (im->has_alpha)
        clean_alpha(loaded_image);
#endif

    create_maps(im);

    /* If there is a segfault it won't be because of the loading. */
    set_loading(NULL);

    return im;
}
