/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 *
 * Guillaume Chazarain <gfc@altern.org>
 */

/***********************
 * Texture management. *
 ***********************/

#include "gliv.h"
#include "math_floats.h"        /* ceilf() */

#include <GL/gl.h>

extern rt_struct *rt;
extern options_struct *options;
extern GtkWidget *gl_widget;

/*
 * Lowest power of two bigger than the argument and 64.
 * glTexImage2D(3G): All implementations support texture images
 * that are at least 64 texels (wide|high).
 */
G_GNUC_CONST static gint gl_dim(gint p)
{
    gint ret = 64;

    while (ret < p)
        ret += ret;

    return ret;
}

/* Draws a piece of a multi-textures image or a whole mono-texture one. */
static void draw_rectangle(gfloat tex_x0, gfloat tex_x1,
                           gfloat tex_y0, gfloat tex_y1,
                           gfloat vert_x0, gfloat vert_x1,
                           gfloat vert_y0, gfloat vert_y1)
{
    /*
     * (tex_x0  ; tex_y0)  : Origin of the interesting part of the texture.
     * (tex_x1  ; tex_y1)  : Extremity of the interesting part of the texture.
     * (vert_x0 ; vert_y0) : Origin of the rectangle.
     * (vert_x1 ; vert_y1) : Extremity of the rectangle.
     */

    glBegin(GL_QUADS);

    glTexCoord2f(tex_x0, tex_y0);
    glVertex2f(vert_x0, vert_y0);

    glTexCoord2f(tex_x1, tex_y0);
    glVertex2f(vert_x1, vert_y0);

    glTexCoord2f(tex_x1, tex_y1);
    glVertex2f(vert_x1, vert_y1);

    glTexCoord2f(tex_x0, tex_y1);
    glVertex2f(vert_x0, vert_y1);

    glEnd();
}

/*
 * Called twice for each rectangle to get rectangle
 * coordinates to put in the display list.
 */
static void compute_coordinates(texture_map * map,
                                gfloat * vert0, gfloat * vert1,
                                gfloat * tex0, gfloat * tex1,
                                gboolean is_x, gint id)
{
    gint map_dim;
    gboolean last;

    if (is_x) {
        map_dim = map->map_init->width;
        last = (id == map->map_init->x_tiles - 1);
    } else {
        map_dim = map->map_init->height;
        last = (id == map->map_init->y_tiles - 1);
    }

    if (id == 0) {
        *vert0 = -map_dim / 2.0;
        *tex0 = 0.0;

        if (last) {
            /* Single tile. */
            *vert1 = -*vert0;
            *tex1 = (gfloat) map_dim / gl_dim(map_dim);
        } else {
            /*
             * First tile.
             * - 1.0: there is only one overlapping pixel.
             */
            *vert1 = *vert0 + rt->max_texture_size - 1.0;
            *tex1 = (rt->max_texture_size - 1.0) / rt->max_texture_size;
        }
    } else {
        /*
         * - 2.0: there are two overlapping pixels.
         * + 1.0: in the first tile there is only one.
         */
        *vert0 = -map_dim / 2.0 + id * (rt->max_texture_size - 2.0) + 1.0;

        if (last) {
            /* Last tile. */
            gint tile_dim;
            gfloat gl_tile_dim;

            tile_dim = map_dim - id * (rt->max_texture_size - 2);
            gl_tile_dim = (gfloat) gl_dim(tile_dim);

            *vert1 = map_dim / 2.0;
            *tex0 = 1.0 / gl_tile_dim;
            *tex1 = tile_dim / gl_tile_dim;
        } else {
            /* Middle tiles. */
            *vert1 = *vert0 + rt->max_texture_size - 2.0;
            *tex0 = 1.0 / rt->max_texture_size;
            *tex1 = 1.0 - *tex0;
        }
    }
}

static void rectangle(texture_map * map, tile_dim * tile, gint i, gint j,
                      gint level)
{
    gfloat ty0, ty1, tx0, tx1;
    gfloat x0, x1, y0, y1;
    gfloat mipmap_coeff;

    mipmap_coeff = pow(MIPMAP_RATIO, -level);

    compute_coordinates(map, &x0, &x1, &tx0, &tx1, TRUE, i);
    compute_coordinates(map, &y0, &y1, &ty0, &ty1, FALSE, j);

    x0 *= mipmap_coeff;
    x1 *= mipmap_coeff;
    y0 *= mipmap_coeff;
    y1 *= mipmap_coeff;

    draw_rectangle(tx0, tx1, ty0, ty1, x0, x1, y0, y1);

    /* Used when drawing, to know which tiles are hidden. */
    tile->x0 = x0;
    tile->y0 = y0;
    tile->x1 = x1;
    tile->y1 = y1;
}

/* Shortcut to OpenGL parameters common to all textures. */
static void texture_parameter(void)
{
    /* These filters will be changed but it's important to initialize them. */
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
}

static void make_texture(gliv_image * im, gint level, gint i, gint j)
{
    static GdkPixbuf *tile_max_size = NULL;
    texture_map *map;
    gboolean last_tile;
    GdkPixbuf *tile;
    gint x, y;
    gint w, h, gl_w, gl_h;

    if (tile_max_size == NULL &&
        im->width >= rt->max_texture_size && im->height >= rt->max_texture_size)

        if (im->width != gl_dim(im->width) ||
            im->height != gl_dim(im->height) ||
            im->width > rt->max_texture_size ||
            im->height > rt->max_texture_size)
            /* First tile. */
            tile_max_size = gdk_pixbuf_new(GDK_COLORSPACE_RGB, im->has_alpha, 8,
                                           rt->max_texture_size,
                                           rt->max_texture_size);

    map = im->maps + level;

    x = i * (rt->max_texture_size - 2);
    y = j * (rt->max_texture_size - 2);

    if (i + 1 == map->map_init->x_tiles) {
        /* Last horizontal tile in this level. */
        w = map->map_init->width - x;
        gl_w = gl_dim(w);

        /* Last level? */
        last_tile = (level + 1 == im->nb_maps);
    } else {
        w = gl_w = rt->max_texture_size;
        last_tile = FALSE;
    }

    if (j + 1 == map->map_init->y_tiles) {
        /* Last vertical tile in this level. */
        h = map->map_init->height - y;
        gl_h = gl_dim(h);
    } else {
        h = gl_h = rt->max_texture_size;
        last_tile = FALSE;
    }

    texture_parameter();

    if (i == 0 && j == 0 && last_tile && w == gl_w && h == gl_h)
        /* Special case, the image is OK for an OpenGL texture. */
        tile = map->map_init->pixbuf;
    else {
        if (gl_w == rt->max_texture_size && gl_h == rt->max_texture_size &&
            tile_max_size != NULL)

            tile = tile_max_size;
        else
            tile = gdk_pixbuf_new(GDK_COLORSPACE_RGB, im->has_alpha, 8,
                                  gl_w, gl_h);

        gdk_pixbuf_copy_area(map->map_init->pixbuf, x, y, w, h, tile, 0, 0);
    }

    if (gl_w != w)
        /* Right border: copy the last column. */
        gdk_pixbuf_copy_area(tile, w - 1, 0, 1, h, tile, w, 0);

    if (gl_h != h) {
        /* Lower corner: copy the last line. */
        gdk_pixbuf_copy_area(tile, 0, h - 1, w, 1, tile, 0, h);

        if (gl_w != w)
            /* Lower-right corner: copy the last pixel. */
            gdk_pixbuf_copy_area(tile, w - 1, h - 1, 1, 1, tile, w, h);
    }

    glTexImage2D(GL_TEXTURE_2D, 0, 3 + im->has_alpha, gl_w, gl_h, 0,
                 im->has_alpha ? GL_RGBA : GL_RGB, GL_UNSIGNED_BYTE,
                 gdk_pixbuf_get_pixels(tile));

    if (tile != tile_max_size && tile != map->map_init->pixbuf)
        gdk_pixbuf_unref(tile);

    if (last_tile && tile_max_size != NULL) {
        gdk_pixbuf_unref(tile_max_size);
        tile_max_size = NULL;
    }
}

static void tiles_parameters(texture_map * map, gboolean is_x)
{
    gint *tiles;
    gint dim;

    if (is_x) {
        tiles = &map->map_init->x_tiles;
        dim = map->map_init->width;
    } else {
        tiles = &map->map_init->y_tiles;
        dim = map->map_init->height;
    }

    *tiles = 1;
    if (dim > rt->max_texture_size)
        *tiles += (gint) ceilf((gfloat) (dim - rt->max_texture_size) /
                               (rt->max_texture_size - 2.0));
}

void prioritize_textures(gliv_image * im, gboolean is_prio)
{
    GLclampf *priorities;
    gint i;

    if (im == NULL)
        return;

    /*
     * im->maps[0] has the biggest nb_tiles, so we don't change the
     * priorities array for each map, we simply use the biggest for all.
     */
    priorities = g_new(GLclampf, im->maps[0].nb_tiles);
    for (i = 0; i < im->maps[0].nb_tiles; i++)
        priorities[i] = (gfloat) is_prio;

    for (i = 0; i < im->nb_maps; i++)
        glPrioritizeTextures(im->maps[i].nb_tiles, im->maps[i].tex_ids,
                             priorities);

    g_free(priorities);
}


static void compute_gl_dimensions(texture_map * map)
{
    tiles_parameters(map, TRUE);
    tiles_parameters(map, FALSE);

    map->nb_tiles = map->map_init->x_tiles * map->map_init->y_tiles;

    map->tex_ids = g_new(guint, map->nb_tiles);
    glGenTextures(map->nb_tiles, map->tex_ids);

    map->list = glGenLists(map->nb_tiles);
}

static void create_a_map(gliv_image * im, gint level)
{
    gint i, j;
    gint id = 0;
    texture_map *map;

    map = im->maps + level;
    compute_gl_dimensions(map);

    map->tiles = g_new(tile_dim, map->nb_tiles);

    for (j = 0; j < map->map_init->y_tiles; j++)
        for (i = 0; i < map->map_init->x_tiles; i++) {

            process_events();

            glBindTexture(GL_TEXTURE_2D, map->tex_ids[id]);
            make_texture(im, level, i, j);

            glNewList(map->list + id, GL_COMPILE);

            /* Redundant but need to be in the display list. */
            glBindTexture(GL_TEXTURE_2D, map->tex_ids[id]);

            rectangle(map, map->tiles + id, i, j, level);
            glEndList();

            id++;
        }
}

/*** Dithering ***/

static void copy_alpha(GdkPixbuf * dest, GdkPixbuf * src)
{
    guchar *dest_data, *src_data, *max_dest;

    process_events();

    src_data = gdk_pixbuf_get_pixels(src);
    dest_data = gdk_pixbuf_get_pixels(dest);
    max_dest = dest_data +
        gdk_pixbuf_get_height(dest) * gdk_pixbuf_get_rowstride(dest);

    while (dest_data < max_dest) {
        dest_data[3] = src_data[3];
        dest_data += 4;
        src_data += 4;
    }
}

/* Returns the original pixbuf if not asked to be destroyed. */
static GdkPixbuf *dither_pixbuf(texture_map * map, gboolean destroy_old)
{
    GdkPixmap *pmap;
    GdkPixbuf *old, *new;
    GdkGC *gc;
    gboolean has_alpha;
    gint width, height;

    gdk_rgb_init();
    process_events();

    old = map->map_init->pixbuf;
    has_alpha = gdk_pixbuf_get_has_alpha(old);
    width = map->map_init->width;
    height = map->map_init->height;

    pmap = gdk_pixmap_new(gl_widget->window, width, height, -1);

    gc = gdk_gc_new(gl_widget->window);

    if (has_alpha)
        gdk_draw_rgb_32_image(pmap, gc, 0, 0, width, height,
                              GDK_RGB_DITHER_MAX, gdk_pixbuf_get_pixels(old),
                              gdk_pixbuf_get_rowstride(old));
    else
        gdk_pixbuf_render_to_drawable(old, pmap, gc, 0, 0, 0, 0, width, height,
                                      GDK_RGB_DITHER_MAX, 0, 0);

    if (has_alpha || destroy_old == FALSE)
        /* We have to keep the old image. */
        new = gdk_pixbuf_new(GDK_COLORSPACE_RGB, has_alpha, 8, width, height);
    else
        new = old;

    process_events();
    gdk_gc_unref(gc);

    gdk_pixbuf_get_from_drawable(new, pmap, gdk_colormap_get_system(),
                                 0, 0, 0, 0, width, height);

    gdk_pixmap_unref(pmap);

    if (has_alpha) {
        copy_alpha(new, old);

        if (destroy_old)
            gdk_pixbuf_unref(old);
    }

    map->map_init->pixbuf = new;
    return destroy_old ? NULL : old;
}

/* Returns the original pixbuf. */
static GdkPixbuf *create_first_map(gliv_image * im)
{
    texture_map *map;
    GdkPixbuf *orig;

    map = im->maps;
    map->map_init->width = im->width;
    map->map_init->height = im->height;

    if (options->dither)
        /* We keep the pixbuf only if mipmaps are used. */
        orig = dither_pixbuf(map, (im->nb_maps == 1));
    else
        orig = map->map_init->pixbuf;

    create_a_map(im, 0);

    if (options->dither)
        gdk_pixbuf_unref(map->map_init->pixbuf);

    g_free(map->map_init);
    return orig;
}

void create_maps(gliv_image * im)
{
    gint level;
    texture_map *map;
    GdkPixbuf *previous;
    gint width, height;

    previous = create_first_map(im);

    if (im->nb_maps == 1) {
        /* No mipmaps. */

        if (options->dither == FALSE)
            gdk_pixbuf_unref(previous);

        prioritize_textures(im, FALSE);
        return;
    }

    width = im->width;
    height = im->height;

    for (level = 1; level < im->nb_maps; level++) {

        map = im->maps + level;

        map->map_init = g_new(loading_data, 1);

        width = map->map_init->width = width * MIPMAP_RATIO;
        height = map->map_init->height = height * MIPMAP_RATIO;
        map->map_init->pixbuf = gdk_pixbuf_scale_simple(previous, width, height,
                                                        GDK_INTERP_BILINEAR);
        gdk_pixbuf_unref(previous);

        if (options->dither)
            /*
             * We must dither each map since rescaling does not preserve it.
             * We keep the original image unless for the last level.
             */
            previous = dither_pixbuf(map, (level == im->nb_maps - 1));

        create_a_map(im, level);

        if (options->dither)
            gdk_pixbuf_unref(map->map_init->pixbuf);
        else
            previous = map->map_init->pixbuf;

        g_free(map->map_init);
    }

    if (options->dither == FALSE)
        gdk_pixbuf_unref(previous);

    prioritize_textures(im, FALSE);
}
