/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 *
 * Guillaume Chazarain <gfc@altern.org>
 */

/**************************
 * The top level widgets. *
 **************************/

#include "gliv.h"

#include <gdk/gdkx.h>           /* gdk_net_wm_supports() */

extern rt_struct *rt;
extern options_struct *options;
extern gliv_image *current_image;
extern GtkWidget *gl_widget;
extern GtkMenuBar *menu_bar;

/* What is in the window. */
static GtkVBox *widgets;

static GtkWindow *main_window;
static GtkWindow *fs_window;

/* The main status bar and its subdivisions. */
static GtkHBox *status_bar;
static GtkEntry *entry_ident;
static GtkEntry *entry_size;
static GtkEntry *entry_state;

static GtkHScrollbar *hscroll;
static GtkVScrollbar *vscroll;

static gboolean window_resized = FALSE;
static gint last_width = 0;
static gint last_height = 0;

/*
 * We determine whether the window has been manually resized in order
 * to prevent automatic resizing in this case.
 */
static gboolean check_window_resized(void)
{
    gint current_width, current_height;

    if (window_resized)
        return TRUE;

    process_events();
    gtk_window_get_size(main_window, &current_width, &current_height);

    window_resized = last_width != 0 &&
        (last_width != current_width || last_height != current_height);

    return window_resized;
}

/*
 * Called when resizing the window because the image changed or a
 * widget has been toggled.
 */
static void resize_window(gint width, gint height)
{
    gtk_window_resize(main_window, width, height);

    last_width = width;
    last_height = height;

    process_events();
}

void show_dialog(GtkWidget * widget, const gchar * name)
{
    /* In windowed mode we let the WM choose the position. */
    if (options->fullscreen)
        gtk_window_set_position(GTK_WINDOW(widget), GTK_WIN_POS_MOUSE);

    if (name != NULL)
        gtk_window_set_title(GTK_WINDOW(widget), name);

    gtk_widget_show_all(widget);
}

static GtkEntry *add_entry(gboolean grow, const gchar * init)
{
    GtkEntry *entry;

    entry = GTK_ENTRY(gtk_entry_new());
    gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
    gtk_box_pack_start(GTK_BOX(status_bar), GTK_WIDGET(entry), grow, TRUE, 0);

    /* We don't want the blinking cursor. */
    GTK_WIDGET_UNSET_FLAGS(GTK_WIDGET(entry), GTK_CAN_FOCUS);

    gtk_entry_set_text(entry, init);

    return entry;
}

static void create_status_bar(void)
{
    status_bar = GTK_HBOX(gtk_hbox_new(FALSE, 0));

    entry_ident = add_entry(TRUE, _("No image loaded"));
    entry_size = add_entry(FALSE, _("width x height"));
    entry_state = add_entry(FALSE, _("zoom% (angle)"));
}

void set_loading_entry(const gchar * filename)
{
    static gchar *old_ident = NULL;

    if (filename == NULL) {
        gtk_entry_set_text(entry_ident, old_ident);
        g_free(old_ident);
    } else {
        gchar *new_ident;

        old_ident = gtk_editable_get_chars(GTK_EDITABLE(entry_ident), 0, -1);

        new_ident = g_strconcat(old_ident, " ", _("loading"), ":",
                                locale_to_utf8(filename), NULL);

        gtk_entry_set_text(entry_ident, new_ident);
        g_free(new_ident);
    }

    if (options->status_bar) {
        /* Make sure the widget is updated now. */
        gtk_widget_queue_draw(GTK_WIDGET(entry_ident));
        gdk_window_process_updates(GTK_WIDGET(entry_ident)->window, TRUE);
    }
}

/* Returns TRUE if the widget has been toggled. */
static gboolean toggle_widget(GtkWidget * widget, gboolean * flag,
                              gboolean horizontal)
{
    gint new_width, new_height, coeff;

    /* We are not reentrant. */
    static gboolean toggling = FALSE;

    if (toggling)
        return FALSE;

    toggling = TRUE;

    process_events();
    gtk_window_get_size(main_window, &new_width, &new_height);

    if (*flag == FALSE) {
        /* Show. */
        gtk_widget_show(widget);
        coeff = 1;
    } else {
        /* Hide. */
        gtk_widget_hide(widget);
        coeff = -1;
    }

    if (horizontal)
        new_width += coeff * widget->allocation.width;
    else
        new_height += coeff * widget->allocation.height;

    *flag ^= TRUE;

    if (options->fullscreen == FALSE)
        resize_window(new_width, new_height);

    toggling = FALSE;
    return TRUE;
}

void toggle_menu_bar(void)
{
    toggle_widget(GTK_WIDGET(menu_bar), &options->menu_bar, FALSE);
}

void toggle_status_bar(void)
{
    toggle_widget(GTK_WIDGET(status_bar), &options->status_bar, FALSE);
}

void toggle_scrollbars(void)
{
    if (!toggle_widget(GTK_WIDGET(hscroll), &options->scrollbars, FALSE))
        return;

    /* To undo the toggle_widget() side effect. */
    options->scrollbars ^= TRUE;

    if (!toggle_widget(GTK_WIDGET(vscroll), &options->scrollbars, TRUE))
        return;

    if (options->scrollbars)
        update_scrollbars(TRUE);
}

void update_status_bar(void)
{
    gchar *size_str, *state_str;
    const gchar *sym;
    gfloat zoom, angle;

    if (current_image == NULL)
        return;

    if (current_image->ident != NULL) {
        /* Filename and dimensions status, only the first time. */
        gtk_entry_set_text(entry_ident, current_image->ident);
        gtk_editable_set_position(GTK_EDITABLE(entry_ident), 0);

        size_str = g_strdup_printf("%dx%d",
                                   current_image->width, current_image->height);
        gtk_entry_set_text(entry_size, size_str);
        g_free(size_str);
    }

    /* The state */

    zoom = get_matrix_zoom() * 100.0;
    angle = get_matrix_angle() * 180.0 / PI;
    sym = is_matrix_symmetry()? " /" : "";

    /* We don't want -0.000 */
    if (float_equal(angle, 0.0))
        angle = 0.0;

    state_str = g_strdup_printf(_("%.3f%% (%.3f deg%s)"), zoom, angle, sym);
    gtk_entry_set_text(entry_state, state_str);

    g_free(state_str);
}

static void update_scroll(GtkScrollbar * scroll, gfloat min, gfloat max,
                          gfloat dim, gboolean redraw)
{
    GtkAdjustment *adj;
    gboolean update_needed;

    adj = gtk_range_get_adjustment(GTK_RANGE(scroll));

    update_needed = (adj->lower < 0.0 || adj->upper > dim ||
                     float_equal(adj->value, 0.0) == FALSE);

    /* smallest length containing the image and the window = window U image. */
    adj->lower = MIN(min, 0.0);
    adj->upper = MAX(max, dim);

    /* window size. */
    adj->page_size = dim;

    /* the offset is actually computed with adj->lower and adj->upper. */
    adj->value = 0.0;

    if (update_needed == FALSE && (adj->lower < 0.0 || adj->upper > dim))
        update_needed = TRUE;

    adj->step_increment = MOVE_OFFSET;
    adj->page_increment = dim;

    gtk_range_set_adjustment(GTK_RANGE(scroll), adj);

    if (options->scrollbars && update_needed && redraw &&
        GTK_WIDGET_REALIZED(GTK_WIDGET(scroll))) {

        gtk_widget_queue_draw(GTK_WIDGET(scroll));
        gdk_window_process_updates(GTK_WIDGET(scroll)->window, FALSE);
    }
}

void update_scrollbars(gboolean redraw)
{
    gfloat min_x, max_x, min_y, max_y;

    if (current_image == NULL)
        return;

    get_matrix_bounding_box(&min_x, &max_x, &min_y, &max_y);

    update_scroll(GTK_SCROLLBAR(hscroll), min_x, max_x,
                  (gfloat) rt->wid_size->width, redraw);

    update_scroll(GTK_SCROLLBAR(vscroll), min_y, max_y,
                  (gfloat) rt->wid_size->height, redraw);
}

static void scroll_val(GtkRange * range)
{
    gfloat offset;

    offset = -1.0 * gtk_range_get_value(range);

    if (range == GTK_RANGE(hscroll))
        matrix_move(offset, 0.0);
    else
        /* range == GTK_RANGE(vscroll) */
        matrix_move(0.0, offset);

    update_scrollbars(FALSE);
    refresh(REFRESH_IMAGE | APPEND_HISTORY);
}

static gboolean has_net_wm(void)
{
    return gdk_net_wm_supports(gdk_atom_intern("_NET_WM_STATE_FULLSCREEN",
                                               FALSE));
}

static GtkRange *scroll_new(gboolean horiz)
{
    GtkRange *range;

    if (horiz)
        range = GTK_RANGE(gtk_hscrollbar_new(NULL));
    else
        range = GTK_RANGE(gtk_vscrollbar_new(NULL));

    gtk_range_set_update_policy(range, GTK_UPDATE_CONTINUOUS);

    g_signal_connect(range, "value-changed", G_CALLBACK(scroll_val), NULL);

    return range;
}

void refresh_all(void)
{
    gtk_widget_queue_draw(GTK_WIDGET(widgets));
    gdk_window_process_updates(GTK_WIDGET(widgets)->window, TRUE);
    process_events();
}

static gboolean first_map(void)
{
    if (options->menu_bar == FALSE)
        gtk_widget_hide(GTK_WIDGET(menu_bar));

    if (options->status_bar == FALSE)
        gtk_widget_hide(GTK_WIDGET(status_bar));

    if (options->scrollbars == FALSE) {
        gtk_widget_hide(GTK_WIDGET(hscroll));
        gtk_widget_hide(GTK_WIDGET(vscroll));
    }

    if (options->fullscreen && has_net_wm())
        toggle_fullscreen(TRUE);

    return FALSE;
}

void create_windows(void)
{
    GtkWindow *first_window;
    GtkAccelGroup *accel_group;
    GtkHBox *hbox;

    /*
     * The main window: seen when not in fullscreen,
     * or when using the NET WM protocol.
     */
    main_window = GTK_WINDOW(gtk_window_new(GTK_WINDOW_TOPLEVEL));
    gtk_window_set_title(main_window, "GLiv");

    install_callbacks(main_window);

    /* The window seen in fullscreen mode. */
    fs_window = GTK_WINDOW(gtk_window_new(GTK_WINDOW_TOPLEVEL));
    gtk_window_set_title(fs_window, _("GLiv in fullscreen"));

    gtk_window_move(fs_window, 0, 0);
    gtk_window_set_default_size(fs_window, rt->scr_width, rt->scr_height);
    gtk_window_set_decorated(fs_window, FALSE);
    install_callbacks(fs_window);

    /* The widgets. */
    create_gl_widget();
    accel_group = create_menus();
    create_status_bar();

    hscroll = GTK_HSCROLLBAR(scroll_new(TRUE));
    vscroll = GTK_VSCROLLBAR(scroll_new(FALSE));

    /* Bind keyboard accelerators to windows. */
    gtk_window_add_accel_group(main_window, accel_group);
    gtk_window_add_accel_group(fs_window, accel_group);

    /* Collection. */
    widgets = GTK_VBOX(gtk_vbox_new(FALSE, 0));
    gtk_box_pack_start(GTK_BOX(widgets), GTK_WIDGET(menu_bar), FALSE, FALSE, 0);

    hbox = GTK_HBOX(gtk_hbox_new(FALSE, 0));
    gtk_box_pack_start_defaults(GTK_BOX(hbox), gl_widget);
    gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(vscroll), FALSE, FALSE, 0);

    gtk_box_pack_start_defaults(GTK_BOX(widgets), GTK_WIDGET(hbox));
    gtk_box_pack_start(GTK_BOX(widgets), GTK_WIDGET(hscroll), FALSE, FALSE, 0);
    gtk_box_pack_end(GTK_BOX(widgets), GTK_WIDGET(status_bar), FALSE, FALSE, 0);

    first_window = (options->fullscreen && !has_net_wm())?
        fs_window : main_window;

    gtk_container_add(GTK_CONTAINER(first_window), GTK_WIDGET(widgets));
    g_signal_connect_after(first_window, "map", G_CALLBACK(first_map), NULL);
    gtk_widget_show_all(GTK_WIDGET(first_window));
}

void goto_window(gboolean force_resize)
{
    gint new_width, new_height;

    if (current_image == NULL)
        return;

    if (force_resize)
        window_resized = FALSE;

    if (force_resize == FALSE && check_window_resized())
        return;

    new_width = current_image->width;
    new_height = current_image->height;

    if (options->menu_bar) {
        new_height += GTK_WIDGET(menu_bar)->requisition.height;
        if (new_width < GTK_WIDGET(menu_bar)->requisition.width)
            new_width = GTK_WIDGET(menu_bar)->requisition.width;
    }

    if (options->status_bar) {
        new_height += GTK_WIDGET(status_bar)->requisition.height;
        if (new_width < GTK_WIDGET(status_bar)->requisition.width)
            new_width = GTK_WIDGET(status_bar)->requisition.width;
    }

    if (options->scrollbars) {
        new_height += GTK_WIDGET(hscroll)->requisition.height;
        new_width += GTK_WIDGET(vscroll)->requisition.width;
    }

    if (new_width >= rt->scr_width || new_height >= rt->scr_height) {
        new_width = 3 * rt->scr_width / 4;
        new_height = 3 * rt->scr_height / 4;
    }

    resize_window(new_width, new_height);
}

static void reparent(GtkWidget * new_parent)
{
    /* Be sure all widgets are displayed. */

    gtk_widget_show(GTK_WIDGET(menu_bar));

    if (options->status_bar && GTK_WIDGET_REALIZED(status_bar) == FALSE)
        gtk_container_foreach(GTK_CONTAINER(status_bar),
                              (GtkCallback) gtk_widget_realize, NULL);

    gtk_widget_show(GTK_WIDGET(status_bar));

    gtk_widget_show(GTK_WIDGET(hscroll));
    gtk_widget_show(GTK_WIDGET(vscroll));

    process_events();

    /* Then reparent them. */
    gtk_widget_reparent(GTK_WIDGET(widgets), new_parent);

    process_events();

    /* Finally, hide widgets previously hidden. */

    if (options->menu_bar == FALSE)
        gtk_widget_hide(GTK_WIDGET(menu_bar));

    if (options->status_bar == FALSE)
        gtk_widget_hide(GTK_WIDGET(status_bar));

    if (options->scrollbars == FALSE) {
        gtk_widget_hide(GTK_WIDGET(hscroll));
        gtk_widget_hide(GTK_WIDGET(vscroll));
    }
}

static gboolean net_wm_toggle_fullscreen(gboolean enable)
{
    if (get_current_window() != main_window || has_net_wm() == FALSE)
        return FALSE;

    if (enable) {
        gtk_window_fullscreen(main_window);
        gdk_window_raise(GTK_WIDGET(main_window)->window);
    } else
        gtk_window_unfullscreen(main_window);

    return TRUE;
}

void toggle_fullscreen(gboolean enable)
{
    /* We first try using the NET WM protocol, then with the old method. */
    if (net_wm_toggle_fullscreen(enable) == FALSE) {
        GtkWindow *new, *old;

        if (enable) {
            /* Go to fullscreen mode. */
            new = fs_window;
            old = main_window;
        } else {
            /* Go to window mode. */
            new = main_window;
            old = fs_window;
        }

        gtk_widget_show(GTK_WIDGET(new));
        gtk_widget_hide(GTK_WIDGET(old));

        reparent(GTK_WIDGET(new));
    }

    if (enable == FALSE)
        goto_window(TRUE);

    options->fullscreen = enable;
    schedule_hide_cursor();
}

void update_window_title(void)
{
    const gchar *title;

    title = locale_to_utf8(current_image->node->data);
    gtk_window_set_title(main_window, title);
}

GtkWindow *get_current_window(void)
{
    return GTK_WINDOW(gtk_widget_get_toplevel(gl_widget));
}
